#!/usr/bin/env python
"""
Open Leaping Lemmings
Lemmings fall from the sky, and go "splat" if you fail to catch them
(C) Copyright 2004 James Paige & Hamster Republic Productions

This file is part of Open Leaping Lemmings.
http://HamsterRepublic.com/games/leaping/

Open Leaping Lemmings is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Open Leaping Lemmings is distributed in the hope that it will be fun,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Open Leaping Lemmings in the file license.txt ;
if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

########################################################################

# standard libraries
import os, sys
from random import random,randint
import ConfigParser
# additional libraries
import pygame
from pygame.locals import *
# local libraries
import loader,textsprite

########################################################################

class Storage:
    """An empty storage class"""

########################################################################

class Fishermen(pygame.sprite.Sprite):
    """A pair of heroic Norwegian fishermen, Birger and Roald, motivated
       by mindless altruism and determined to save as many poor foolish
       lemmings as possible"""
    def __init__(self,game):
        # descended from pygame sprite
        pygame.sprite.Sprite.__init__(self)
        # parent game reference
        self.game = game
        # images
        self.imgUp = game.loader.surface('fishermen-up.png')
        self.imgDown = game.loader.surface('fishermen-down.png')
        self.imgLeft = game.loader.surface('fishermen-left.png')
        self.imgRight = game.loader.surface('fishermen-right.png')
        self.image = self.imgUp
        self.rect = self.image.get_rect()
        # movement tracking
        self.go = 0.0
        self.wantRight = 0
        self.wantLeft = 0
        # net hit-detection rectangle
        self.netRect = Rect((0,0,36,8))
        # net bounce animation tracker
        self.bounceFrames = 0

    def __del__(self):
        self.game = None

    def update(self):
        # restrict motion off the edges of the playing field
        if self.netRect.left < self.game.beach.rect.left:
           self.wantLeft = 0
        if self.netRect.right > self.game.beach.rect.right:
           self.wantRight = 0
        # where do we currently want to go?
        self.go += self.wantRight
        self.go -= self.wantLeft
        # make sure we don't go too fast
        if self.go > 5.5: self.go = 5.5
        if self.go < -5.5: self.go = -5.5
        # actually move the fisherment
        self.rect.move_ip((self.go,0))
        self.netRect.center = self.rect.center
        self.netRect.move_ip(3,16)
        # cause go to decay
        if self.go > 0:
            self.go -= .5
            if self.go < 0: self.go = 0
        if self.go < 0:
            self.go += .5
            if self.go > 0: self.go = 0
        # recoil bounce animation
        if self.bounceFrames > 0:
            self.bounceFrames -= 1
            if self.bounceFrames == 0:
                self.image = self.imgUp

    def bounce(self,x = 0):
        """animate something bouncing off the net"""
        # see which side of the net the bounce occured on
        xOffset = x - self.netRect.centerx
        # set the number of frames to animate
        self.bounceFrames = 4
        # change to a bounce pic
        if xOffset < -9:
            self.image = self.imgLeft
        elif xOffset > 9:
            self.image = self.imgRight
        else:
            self.image = self.imgDown

########################################################################

class LemmingHerd:
    """the source of the lemmings who keep falling off the cliff"""
    def __init__(self,game):
        self.game = game
        # load available lemming graphics
        self.images = []
        for i in range(1,6):
            closedEyes = game.loader.surface('lemming-%dc.png' % i)
            openEyes = game.loader.surface('lemming-%do.png' % i)
            self.images.append((closedEyes,openEyes))
        # load available lemming splat graphics
        self.splatImages = []
        for i in range(1,8):
            self.splatImages.append(game.loader.surface('splat-%d.png' % i))
        # create recycling bin for lemmings and splats
        self.recycle = Storage()
        self.recycle.lemming = []
        self.recycle.splat = []

    def __del__(self):
        self.game = None

    def update(self):
        if randint(0,999) >= 960:
            self.spawnLemming()

    def spawnLemming(self):
        """recycles a dead lemming to the top of the cliff, or spawns a
           new one if no recyclable dead lemming ghosts are available"""
        if len(self.recycle.lemming):
            lem = self.recycle.lemming[0]
            self.recycle.lemming.pop(0)
        else:
            lem = Lemming(self.game,self.images[randint(0,len(self.images)-1)])
        lem.startFall()
        self.game.score.fallen.val += 1
        self.game.score.update()

    def spawnSplat(self, center):
        """recycles a finished splat from the splat bin, or spawns a new
           splat if no recycleable finished splats are available"""
        if len(self.recycle.splat):
            splat = self.recycle.splat[0]
            self.recycle.splat.pop(0)
        else:
            splat = LemmingSplat(self.game,center)
        splat.startSplat(center)
        self.game.visibleSprites.add(splat)

########################################################################

class Lemming(pygame.sprite.Sprite):
    """A foolish Lemming, who falls to his (possibly avertable) death"""
    def __init__(self, game, lemmingImagePair):
        pygame.sprite.Sprite.__init__(self)
        self.game = game
        (self.eyesClosed,self.eyesOpen) = lemmingImagePair
        self.image = self.eyesClosed
        self.rect = self.image.get_rect()

    def __del__(self):
        self.game = None

    def update(self):
        """causes the Lemming to fall, in accordance with the rules of gravity"""
        #gravity
        self.fallSpeed += self.game.physics.gravity
        self.fallSpeed *= 1-self.game.physics.wind_drag
        # fall
        self.fallTracker += self.fallSpeed
        fallInt = int(self.fallTracker)
        self.fallTracker -= fallInt
        # apply movement
        self.rect.move_ip(0,fallInt)
        # keep lemming on the playing area
        if self.rect.left < self.game.beach.rect.left:
            self.rect.left = self.game.beach.rect.left
        if self.rect.right > self.game.beach.rect.right:
            self.rect.right = self.game.beach.rect.right
        # check to see if the lemming has hit the sand
        if self.rect.bottom >= self.game.screenRect.height:
            self.hitSand()
        # check to see if the lemming has hit the net
        elif self.fallSpeed >= 0 and self.bounces < 2 and self.rect.colliderect(self.game.fishermen.netRect):
            self.hitNet()

    def hitNet(self):
        # animate the net
        self.game.fishermen.bounce(self.rect.centerx)
        # bounce
        self.fallSpeed = -self.fallSpeed
        # apply recoil modifier
        if self.bounces == 0:
            self.fallSpeed *= self.game.physics.first_bounce
            self.game.score.bounced.val += 1
            # animate lemming's eyes
            self.image = self.eyesOpen
        else:
            self.fallSpeed *= self.game.physics.subsequent_bounces
        # play bounce sound
        self.game.sound_bounce.play()
        # increment bounce counter
        self.bounces += 1
        # update scores
        self.game.score.update()

    def hitSand(self):
        """when the lemming hits the sand, it either dies or is saved
           depending on how many times it has bounced"""
        if self.bounces >=2:
            self.game.score.saved.val += 1
        else:
            self.game.score.died.val += 1
            self.game.sound_fall[randint(0,len(self.game.sound_fall)-1)].play()
            self.game.herd.spawnSplat(self.rect.center)
        self.game.score.update()
        self.game.visibleSprites.remove(self)
        self.game.herd.recycle.lemming.append(self)

    def startFall(self):
        # randomly rotate images
        rotate = randint(1,3) * 90
        flipX = randint(0,1)
        flipY = randint(0,1)
        self.eyesOpen = pygame.transform.rotate(pygame.transform.flip(self.eyesOpen,flipX,flipY),rotate)
        self.eyesClosed = pygame.transform.rotate(pygame.transform.flip(self.eyesClosed,flipX,flipY),rotate)
        # switch to eyes-closed image
        self.image = self.eyesClosed
        self.rect = self.image.get_rect()
        # move the lemming to the top of the cliff
        playRect = self.game.beach.rect
        self.rect.midtop = (randint(playRect.left + self.rect.width / 2 \
                            ,playRect.left + playRect.width-self.rect.width / 2) \
                            ,-self.rect.height)
        # reset bounce count
        self.bounces = 0
        # set random falling speed
        self.fallSpeed = randint(220,350)/100.0
        self.fallTracker = 0.0
        # add to the visible group
        self.game.visibleSprites.add(self)

########################################################################

class LemmingSplat(pygame.sprite.Sprite):
    """A lemming splats when it hits the sand"""
    def __init__(self, game, center=(0,0)):
        pygame.sprite.Sprite.__init__(self)
        self.game = game
        self.startSplat(center)

    def __del__(self):
        self.game = None

    def startSplat(self,center):
        self.center = center
        self.frame = 0
        self.updateFrame()

    def update(self):
        """causes the splat to cycle through its frames"""
        if self.ticks + 35 < pygame.time.get_ticks():
            if self.frame < 6:
                self.frame += 1
                self.updateFrame()
            else:
                self.game.visibleSprites.remove(self)
                self.game.herd.recycle.splat.append(self)
            
    def updateFrame(self):
        self.image = self.game.herd.splatImages[self.frame]
        self.rect = self.image.get_rect()
        self.rect.center = self.center
        self.ticks = pygame.time.get_ticks()


########################################################################

class Scores:

    def __init__(self, font):
        scoreLeft = 5
        spacing = 1
        # fallen
        self.fallen = Score(font,0,color=(60,60,0),topleft=(scoreLeft,0))
        # died
        self.died = Score(font,0,color=(100,0,0),topleft=(scoreLeft,self.fallen.sprite.rect.bottom+spacing))
        # bounced
        self.bounced = Score(font,0,color=(0,0,100),topleft=(scoreLeft,self.died.sprite.rect.bottom+spacing))
        # saved
        self.saved = Score(font,0,color=(0,100,0),topleft=(scoreLeft,self.bounced.sprite.rect.bottom+spacing))
        # chart
        self.chart = pygame.sprite.Sprite()
        self.chart.image = pygame.Surface((51,160))
        self.chart.rect = self.chart.image.get_rect()
        self.chart.rect.topleft = (scoreLeft,self.saved.sprite.rect.bottom+spacing+2)
        # time
        self.time = Score(font,color=(60,60,60),topleft=(scoreLeft,self.chart.rect.bottom+spacing))
        # update
        self.update()

    def update(self):
        self.fallen.sprite.text  = 'Fallen: %d' % self.fallen.val
        self.died.sprite.text    = 'Died: %d' % self.died.val
        self.bounced.sprite.text = 'Bounced: %d' % self.bounced.val
        self.saved.sprite.text   = 'Saved: %d' % self.saved.val
        self.time.sprite.text    = 'Time: %02d:%02d:%02d' % self.splitTime(self.time.val)
        #self.time.sprite.text    = 'Time: %d' % self.time.val
        self.drawChart()

    def splitTime(self,time):
        hours = self.time.val / 3600
        minutes = self.time.val % 3600 / 60
        seconds = self.time.val % 60
        return (hours,minutes,seconds)
        
    def sprites(self):
        return self.chart, self.fallen.sprite, self.died.sprite, self.bounced.sprite, self.saved.sprite, self.time.sprite

    def drawChart(self):
        # a list of the scores we want to chart
        chartables = (self.died,self.bounced,self.saved)
        # init a target rectangle the same size as the chart, but at position 0,0
        targetRect = Rect(self.chart.rect)
        targetRect.topleft = (0,0)
        # lock the chart for direct drawing
        self.chart.image.lock()
        # draw a background rect over the whole chart
        pygame.draw.rect(self.chart.image,self.fallen.color,targetRect)
        # shrink the target rect to make a one pixel border
        targetRect.inflate_ip(-2,-2)
        targetRect.left += 1
        # adjust the target rect width for drawing one meter.
        # It must be the right size to allow all meters to fit
        targetRect.width = (targetRect.width / len(chartables)) - 1
        if self.fallen.val:
            for score in chartables:
                newSize = (targetRect.height / float(self.fallen.val)) * score.val
                if newSize:
                    meterRect = Rect(targetRect)
                    meterRect.height = newSize
                    meterRect.bottom = targetRect.bottom-1
                    pygame.draw.rect(self.chart.image,score.color,meterRect)
                targetRect.left += targetRect.width+1
        # unlock the chart after direct drawing is finished
        self.chart.image.unlock()
        
########################################################################

class Score:

    def __init__(self,font,startValue=0,color=(128,128,128),topleft=(0,0)):
        self.val = 0
        self.color = color
        self.sprite = textsprite.TextSprite(font,bgcolor=color)
        self.sprite.rect.topleft = topleft

########################################################################

class LemmingGame:

    def __init__(self):
        # init pygame
        pygame.init()
        try:
            pygame.mixer.init()
        except pygame.error:
            print "WARNING: Sound Support disabled: %s" % pygame.get_error()
            pass
        # set game physics
        self.physics = Storage()
        self.physics.gravity = .1
        self.physics.wind_drag = 0.0125
        self.physics.first_bounce = 1.4
        self.physics.subsequent_bounces = 0.4
        # set default user-configurable settings
        self.data = os.path.join(os.path.dirname(sys.argv[0]),'data')
        self.fullscreen = 0
        # load configuration files (which may override the above settings)
        self.readConfig()
        # instantiate data file reader
        self.loader = loader.DataDirReader(self.data)
        # init joystick
        if pygame.joystick.get_count():
            self.joy = pygame.joystick.Joystick(0)
            self.joy.init()
        # init screen mode, create window
        self.screen = pygame.display.set_mode((640,480),DOUBLEBUF|self.fullscreen)
        pygame.display.set_caption('Leaping Lemmings')
        icon = self.loader.surface('lemming-icon.png')
        pygame.display.set_icon(icon)
        self.screenRect = self.screen.get_rect()
        pygame.mouse.set_visible(0)
        # load font
        self.font = self.loader.font('verasans.ttf',12)
        # create a clock for timing control
        self.clock = pygame.time.Clock()
        # load and position background
        self.beach = self.loader.sprite('cliff-narrow.png')
        self.beach.rect.center = self.screenRect.center
        # create game elements
        self.score = Scores(self.font)
        self.scoreSprites = pygame.sprite.RenderPlain(self.score.sprites())
        self.visibleSprites = pygame.sprite.RenderPlain()
        # create fishermen
        self.fishermen = Fishermen(self)
        self.fishermen.rect.midbottom = self.beach.rect.midbottom
        self.fishermen.rect.bottom -= 8
        self.visibleSprites.add(self.fishermen)
        # create Lemming herd
        self.herd = LemmingHerd(self)
        # load sound effects
        self.sound_fall = []
        for sound_file in self.loader.file_regex('fall-.*\.ogg'):
            self.sound_fall.append(self.loader.sound(sound_file))
        self.sound_bounce = self.loader.sound('bounce.ogg')   
        # paint the first frame
        self.paint()
        # reset the play timer
        self.clock.tick()
        self.score.time.val = 0
        self.timeTracker = 0.0
        # start the music
        self.loader.music('mountain-king')

    def __del__(self):
        # clean up pygame
        pygame.quit()
        pygame.mixer.quit()

    def quit(self):
        self.playing = False

    def readConfig(self):
        config = ConfigParser.RawConfigParser()
        detected_os = os.name
        if os.name == 'nt':
            # config file location for Windows
            if os.path.expanduser('~') == '~':
                # Windows 9X may not support home directories...
                detected_os = 'win9x'
                # therefore the config files must live in the application dir
                configFile = os.path.join(os.path.dirname(sys.argv[0]),'conf','leaping.conf')
            else:
                # for NT/2K/XP expand the home directory...
                configFile = os.path.join(os.path.expanduser('~'),'Application Data','Hamster Republic','Open Leaping Lemmings','leaping.conf')
        else:
            # config file location for all other OS's (unixen and linuces)
            configFile = os.path.join(os.path.expanduser('~'),'.hamsterrepublic','leaping.conf')
        print "Operating System: %s" % detected_os
        print "Seeking configuration file: %s" % configFile
        if os.access(configFile,os.R_OK):
            # read the configuration file...
            print "Reading configuration file: %s" % configFile
            config.read(configFile)
            # interpret various config settings from the config file.
            if config.has_option('config','gamedata'):
                # an alternate location for the game data files
                dataFile = config.get('config','gamedata')
                if os.access(dataFile,os.R_OK):
                    self.data = dataFile
                else:
                    print "%s: Specified data directory %s is not readable." % (configFile,dataFile)
            if config.has_option('config','fullscreen'):
                # fullscreen or windowed mode
                try:
                    if config.getboolean('config','fullscreen'):
                        self.fullscreen = FULLSCREEN
                    else:
                        self.fullscreen = 0
                except ValueError:
                    print "%s: Invalid setting for fullscreen. Should be true or false." % configFile

    def play(self):
        # main game loop
        self.playing = True
        while self.playing:
            # speed control
            self.countSeconds(self.clock.tick(32))
            # user input
            self.control()
            # game state update
            self.visibleSprites.update()
            self.herd.update()
            # screen update
            self.paint()
        # end of play loop

    def control(self):
        # event handling
        for event in pygame.event.get():
            if event.type == QUIT:
                self.quit()
            elif event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    self.quit()
                elif event.key == K_RETURN and event.mod & KMOD_ALT:
                    self.fullscreen = self.fullscreen ^ FULLSCREEN
                    self.screen = pygame.display.set_mode((640,480),DOUBLEBUF|self.fullscreen)
                elif event.key == K_LEFT:
                    self.fishermen.wantLeft = 1
                elif event.key == K_RIGHT:
                    self.fishermen.wantRight = 1
            elif event.type == KEYUP:
                if event.key == K_LEFT:
                    self.fishermen.wantLeft = 0
                elif event.key == K_RIGHT:
                    self.fishermen.wantRight = 0
            elif event.type == JOYAXISMOTION:
                # we only care about the X axis
                if event.axis == 0:
                   if event.value == 0:
                      self.fishermen.wantRight = 0
                      self.fishermen.wantLeft = 0
                   elif event.value < 0:
                      self.fishermen.wantLeft = 1
                   elif event.value > 0:
                      self.fishermen.wantRight = 1

    def paint(self):
        # paint background
        self.screen.blit(self.beach.image,self.beach.rect)
        # paint sprites
        self.screen.set_clip(self.beach.rect)
        self.visibleSprites.draw(self.screen)
        # paint score
        self.screen.set_clip()
        self.scoreSprites.draw(self.screen)
        # flip double buffer
        pygame.display.flip()

    def countSeconds(self,milliseconds):
        self.timeTracker += milliseconds/1000.0
        if self.timeTracker >= 1:
            self.timeTracker -= 1
            self.score.time.val += 1
            self.score.update()

########################################################################

if __name__ == '__main__':
    # Import Psyco if available
    try:
        import psyco
        psyco.full()
    except ImportError:
        pass
    # instantiate and play a game object
    game = LemmingGame()
    game.play()
