"""
Various utility functions related to loading game data for pygame games
(C) Copyright 2004 James Paige & Hamster Republic Productions

This file is part of Open Leaping Lemmings.
http://HamsterRepublic.com/games/leaping/

Open Leaping Lemmings is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Open Leaping Lemmings is distributed in the hope that it will be fun,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Open Leaping Lemmings in the file license.txt ;
if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

########################################################################

import os
import re
import pygame
from pygame.locals import *

########################################################################

# verify that PNG image loading support is enabled
if not pygame.image.get_extended():
   print 'Error: extended image formats not supported. pygame must be compiled with SDL_image support.'
   raise SystemExit

# verify that fonts are supported
if not pygame.font:
   print 'Error: font support is not available.'
   raise SystemExit

########################################################################

def normalize_filename(unix_filename):
    """takes a filename with unix-style / path separators and returns
    a filename that will work on the current os"""
    return os.path.join(*unix_filename.split("/"))

def surface(filename,alpha=1):
    """load a specified file as a Surface"""
    try:
        image = pygame.image.load(filename)
    except pygame.error, message:
        print 'Error: Cannot load image:', filename
        raise SystemExit, message
    if alpha:
      image = image.convert_alpha()
    else:
      image = image.convert()
    return image

def sprite(filename,alpha=1):
    """loads the specified file as a Sprite"""
    spr = pygame.sprite.Sprite()
    spr.image = surface(filename,alpha)
    spr.rect = spr.image.get_rect()
    return spr

def font(filename, size):
    """loads a ttf font"""
    return pygame.font.Font(filename, size)
    
def music(filename, play=True):
    # autodetect best file type
    if os.access(filename+'.ogg',os.R_OK):
        filename+='.ogg'
    elif os.access(filename+'.mp3',os.R_OK):
        filename+='.mp3'
    elif os.access(filename+'.wav',os.R_OK):
        filename+='.wav'
    elif os.access(filename+'.mod',os.R_OK):
        filename+='.mod'
    elif os.access(filename+'.mid',os.R_OK):
        filename+='.mid'
    else:
        # if no supported filetypes are found, load nothing
        print "music not available: %s.[ogg|mp3|wav|mod|mid]" % filename
        return
            
    # load the file
    try:
        pygame.mixer.music.load(filename)
        if play:
            pygame.mixer.music.play(-1)
    except pygame.error:
        print "unable to load music: %s" % filename

def sound(filename):
    try:
        return pygame.mixer.Sound(filename)
    except pygame.error:
        return DummySound()
        
def file_regex(filename_regex, dir_to_list, prepend_path=False):
    """returns a list of filenames from a folder that match a regex"""
    regex = re.compile(filename_regex)
    result = []
    for filename in os.listdir(dir_to_list):
        match = regex.search(filename)
        if match is not None:
            if prepend_path:
                filename = os.path.join(dir_to_list,filename)
            result.append(filename)
    return result

########################################################################

class DataDirReader:
    """This class encapsulates the same helper functions available
       directly in the loader module, but it also keeps track of a
       chosen data directory and makes all paths relative to it.
       Also, unlike the regular loader funtions it always expects
       its input filenames to use unix-style path separators"""
       
    def __init__(self,dataDir):
        self.dataDir = dataDir

    def fix(self,unix_filename):
        return os.path.join(self.dataDir, normalize_filename(unix_filename))
        
    def surface(self,unix_filename,alpha=1):
        filename = self.fix(unix_filename)
        return surface(filename,alpha)

    def sprite(self,unix_filename,alpha=1):
        filename = self.fix(unix_filename)
        return sprite(filename,alpha)

    def font(self,unix_filename,size):
        filename = self.fix(unix_filename)
        return font(filename,size)

    def music(self,unix_filename,play=True):
        filename = self.fix(unix_filename)
        music(filename)

    def sound(self,unix_filename):
        filename = self.fix(unix_filename)
        return sound(filename)
        
    def file_regex(self, filename_regex, unix_dir_to_list='./', prepend_path=False):
        unix_dir_to_list = self.fix(unix_dir_to_list)
        return file_regex(filename_regex, unix_dir_to_list, prepend_path)

########################################################################

class DummySound:
    """
    A do-nothing sound class for when real sound support is unavailable
    """

    def play(self):
        pass

########################################################################
